# DirSelect: A Tk directory selection widget.
#
# This widget allows navigating MS Windows local and mapped network
# drives and directories and selecting a directory.
#
# Perl/Tk includes several widgets to perform the function of selecting a file or directory. 
# However, widgets originally written  for Unix sysyems (such as FileSelect) don't allow browsing
# local and mapped drives under Windows, and widgets adapted for Win32 systems that do
# allow this (such as GetOpenFile) don't allow the user to select a directory instead of a file.
#
# On non-MS systems, this is simply a dialog box with a Dirtree in it.
# 
# Usage: my $dir = $mainwindow->DirSelect->Show;
#
# Email comments, questions or bug reports to Kristi Thompson, kristi@brainimplants.com


package DirSelect;
use vars qw($VERSION);
$VERSION = '1.0';
# Supersedes all versions of Win32DirSelect

@EXPORT_OK = qw(glob_to_re);
use strict;
require Tk::Derived;
use Tk qw(Ev);
use vars qw(@EXPORT_OK);
use base qw(Tk::Toplevel);
use Tk::widgets qw(Frame Label DirTree);
use Cwd;

Construct Tk::Widget 'DirSelect';

sub Populate {
   my ($cw, $args) = @_;   
   $cw->SUPER::Populate($args);
   
   # create three frames to hold the drive information (on Win32 systems), 
   # the directory tree and the buttons
   my $top = $cw->Frame->pack(-fill=>'x', -expand=>'1',-side=>'top');
   my $bottom = $cw->Frame->pack(-fill=>'x', -expand=>'1',-side=>'bottom');
   my $mid = $cw->Frame->pack(-fill=>'both', -expand=>'1');
   
   # create the OK and Cancel buttons, and attach the appropriate commands
   my $button = $bottom->Button(-width=>'7', 
   	-text=>'Cancel', 
   	-command=>sub{$cw->{'dir'} = undef})->pack(-anchor=>'e', -side =>'right', -pady=>'1m'); # returns undef
   	
   
   my $button = $bottom->Button(-width=>'7',-text=>'OK', -command=>sub{$cw->{'dir'} = $mid->packSlaves->selectionGet()}, -justify=>'right')->pack(-anchor=>'e', -side =>'right', padx=>'1m', -pady=>'1m'); # packSlaves->selectionGet returns the currently selected element of dirtree
   
   # Check the OS environment variable to find out if this is a MSWin32 system or something else, 
   # and create the meat of the widget accordingly  
   
   if (substr($^O, 0, 5) =~ /mswin/i){		# This is a MSWin32 system
      	
   	require Win32API::File;					
   	# Win32API::File contains functions allowing reading and manipulation of Win32 drive information. 
   	
   	my @drives = Win32API::File::getLogicalDrives();	      
   	my $v;
	   my $state;
	   
	   # Create a radio button for each drive on the system, 
	   # and associate it with a command to browse that drive
	   foreach my $d (@drives){				
			my $drive = sprintf "%-2.2s", $d; 	
			no strict 'refs';
			
			# $$d is a soft reference, so turned off strict refs. 
			# The variable name of the button is set to the logical drive name.  
			# This lets me later invoke the button associated with a particular drive. 
			
			$$d=$top->Radiobutton(	-state=>$state,	
				-selectcolor => $top->cget(-background),
				-indicatoron => '0',
				-text => $drive, 
				-command=>[\&Browse, $mid, $d],
				-value=>$d, 
				-variable =>\$v, 
				-width=>'3')->pack(-side=>'left', -padx=>'.5m', -pady=>'1m' );
		}
		
		
		my $currentdir  = cwd;	
		my $startdrive = substr($currentdir, 0, 2);
		
	  	foreach my $d (@drives){
	  		if (substr($d, 0, 2) =~ /$startdrive/i){
	  			no strict 'refs';
	  			# Invoke the appropriate button so DirSelect initially shows the current drive. 
	   		$$d->invoke;				
	   		last;
      	}
   	}
   }
   else {	
   	# Unix system, presumably. No fancy stuff with drives needed, 
   	# so just slap the DirTree into the middle frame.
   	my $d = '/';
   	chdir $d;
   	my $browse = $mid->Scrolled('DirTree', 
   		-directory => $d, 
   		-scrollbars=>'ose', 
   		-bg=>'white')->pack(-side=>'bottom', -fill=>'both', -expand=>'1');
	}

}


sub Browse {

   my ($mid, $d) = @_;

	# Hash storing drive type info displayed above the DirTree.
	# Win32API::File::GetDriveType returns an integer corresponding to the drive type.
	
	my %drives = (	'0' =>'Unknown', '1'=> 'No root drive', 				
	         		'2'=>'Removable disk drive', '3'=>'Fixed disk drive',
	         		'4'=>'Network drive', '5'=>'CD-Rom drive', 
         			'6'=>'RAM Disk');
         		
   my $drivetype= Win32API::File::GetDriveType( $d );
   my $drive = uc $d;
   my @children = $mid->packSlaves;
   foreach my $c (@children){
      $c -> packForget;					# Get rid of previous drive browser so we can make a new one.
   }
   
   unless (chdir $d){
    	$mid->Label(-justify=>'left', -text=>"$drive is not available.")->pack(-anchor=>'w', -side=>'top');	
  	  	return;
   }   
   my $volumelabel;
   
   Win32API::File::GetVolumeInformation( $d, $volumelabel, [], [], [], [], [], [] );
   
   my $drivetype= Win32API::File::GetDriveType( $d );
   my $drive = uc $d;
   
   $mid->Label(-text=>"$volumelabel ($drive) $drives{$drivetype} ")->pack(-anchor=>'w', -side=>'top');
   
   my $browse = $mid->Scrolled('DirTree', 
   	-directory => $d, 
   	-scrollbars=>'ose', 
   	-bg=>'white')->pack(-side=>'bottom', -fill=>'both', -expand=>'1');

}

sub Wait {
   my $cw = shift;
   $cw->waitVariable(\$cw->{'dir'});
   $cw->grabRelease;
   $cw->withdraw;
   $cw->Callback(-command => $cw->{'dir'});
}

sub Show {
   my ($cw, $grab) = @_;
   my $old_focus = $cw->focusSave;
   my $old_grab = $cw->grabSave;
   $cw->Popup();
   Tk::catch {
      if (defined $grab && length $grab && ($grab =~ /global/)) {
	 		$cw->grabGlobal;
      } 
      else {
	 		$cw->grab;
      }
   };
   $cw->focus;
   $cw->Wait;
   &$old_focus;
   &$old_grab;
   return $cw->{'dir'};
}

1;
