

#!/usr/bin/perl -w

# $Id: vroom,v 1.3 2000/02/25 12:00:46 gabrielx Exp gabrielx $
# This file uses width 3 tabs.
#
# Copyright (c) 2000 Gabriel M. Deal.  All rights reserved.
# This program is free software; you can redistribute it and/or modify it
# under the same terms as Perl itself.


=head1 NAME

vroom - a Perl/Tk program for practicing your typing

=head1 SYNOPSIS

vroom [--characters=N] [--competitive] [--help] [--nocompetitive] [--speed=INTEGER] [FILE]

=cut

my $Usage;
{
	my($prog) = ($0 =~ m#([^/]+)$#);
	my $space = " " x length $prog;
	($Usage = <<EOF) =~ s/^\t\t//gm;
		Usage: $prog [--characters=N] [--competitive] [--help] [--nocompetitive] 
		       $space [--speed=INTEGER] [FILE]

		Use ``perldoc $prog'' for more information.
EOF
}

=head1 DESCRIPTION

B<vroom> displays text to type, when you finish typing the text, it
displays your typing speed.   There are two modes: competitive and
non-competitive.   In competitive mode you must keep up with the speed
that B<vroom> is displaying text at, but if you type faster than that
speed, then, like cruise control in cars, it will temporarily speed
up to accomodate you.   The speed at which it presents you with new
text and the mode it runs in are both configurable on the command line
and at runtime. 

B<vroom> uses a normal ASCII file to get the text it displays.   Any
sequence of whitespace (tab, space, or newline) is converted to a single
space because otherwise it is difficult to tell the difference between
the three kinds of whitespace. 

=head1 OPTIONS

=over

=item * --characters=N

the number of characters that will be displayed at once (default is 50).

=item * --competitive

run in competitive mode (default).

=item * --help

display a short usage message.

=item * --nocompetitive

run in non-competitive mode.

=item * --speed=INTEGER

set the speed (in characters per minute) that new characters are
displayed at to INTEGER, only used in competitive mode (default is
60).

=item * FILE

the data file to load initially.

=back

=head1 AUTHOR

gabriel m. deal
<gmd@yellowleaf.org>
http://www.yellowleaf.org/gmd/

=cut

use strict;
use diagnostics;
###############
package Misc;


#=# xdie(;@)
#=# 
#=# Exit by passing the elements of the passed array to xdie(), if the last
#=# element in the passed array does not end with a newline (or no
#=# arguments are passed), also print out a stack trace after printing the
#=# other passed arguments.
#=# 
#=#
sub xdie
{
	if(@_ == 0 or $_[$#_] !~ /\n$/)
	{
		my @stack_trace = stack_trace();
		pop @stack_trace;
		die join(' ', @_)
			. "\nStack Trace:\n\t"
			. join("\n\t", @stack_trace)
			. "\n";
	}
	else
	{
		die join(' ', @_);
	}
}

#=# stack_trace()
#=# 
#=# Return an array containing filename, function, and line information
#=# about each function currently on the stack.
#=# 
#=#
sub stack_trace
{
	my $i = 0;
	my($filename, $line, $subroutine, @stack_trace);
	while(($filename, $line, $subroutine) = (caller $i)[1, 2, 3])
	{
		$i++;
		unshift @stack_trace, "$subroutine() called at line $line of $filename";
	}

	return @stack_trace;
}


1;
###############
package MovingLine;

#=# =head1 NAME
#=# 
#=# MovingLine - manages line of characters parading across a Canvas.
#=# 
#=# =head1 DESCRIPTION
#=# 
#=# A MovingLine is a sequence of characters that place characters on a
#=# Canvas, move them across the Canvas, and remove them from the Canvao.
#=# The movement of the characters is directed by a Curve object.
#=# 
#=# =head1 CONSTRUCTOR
#=# 
#=# =over
#=# 
#=# =item new($curve, $parent) : $MovingLine
#=# 
#=# $curve is a Curve object that will determine the curve the characters
#=# will follow.   $parent is the Canvas object each point will be placed on.
#=# 
#=# =back
#=# 
#=# =cut
sub new
{
	my($class, $curve, $parent) = @_;

	my $self = {
		points => [],
		curve => $curve,
		adispatch => {
			text => $curve->can("forward_point"),
			rectangle => $curve->can("forward_rectangle")
		},
		rdispatch => {
			text => $curve->can("backward_point"),
			rectangle => $curve->can("backward_rectangle")
		},
		parent => $parent
	};

	return bless($self, ref($class) || $class);
}

sub _ispastmark
{
	my($self, $dir, $point) = @_;

	return 0 unless exists $point->{"${dir}mark"};
	($dir eq "a") 
		? 0 > $self->{curve}->cmppt($point->{coord}, $point->{"${dir}mark"})
		: 0 < $self->{curve}->cmppt($point->{coord}, $point->{"${dir}mark"});
}

#=# =head1 METHODS
#=# 
#=# =over
#=# 
#=# =item advance() : -
#=# 
#=# Move each character in this MovingLine forward on the curve.   Remove
#=# the characters from the Canvas when they reach the end of the curve.
#=# 
#=# AFTER moving each character, throw an exception starting with
#=# "REMOVED" if any character reached the end of the curve.
#=# 
#=# =cut
sub advance { $_[0]->_move("a") }
sub retreat { $_[0]->_move("r") }

sub _move
{
	my($self, $dir) = @_;

	my($flag, $mflag) = (0, 0);
	foreach(@{ $self->{points} })
	{
		if($self->_ispastmark($dir, $_))
		{
			$_->{"${dir}cb"}->();
			next;
		}
		eval { 
			$self->{"${dir}dispatch"}{$_->{type}}->($self->{curve}, $_->{coord}) 
		};
		if($@)
		{
			die unless $@ =~ /^END/;
			undef $@;
			$flag++;
			$self->{parent}->delete($_->{obj});
		}
		else { $self->{parent}->coords($_->{obj}, @{ $_->{coord} }) }
	}
	if($dir eq "a") { foreach(1 .. $flag) { CORE::shift(@{ $self->{points} }) }}
	else { foreach(1 .. $flag) { pop(@{ $self->{points} }) } }

	Misc::xdie "REMOVED. " if $flag ;
}

#=# =item DESTROY() : -
#=# 
#=# Remove all characters from the canvas.
#=# 
#=# =cut
sub DESTROY
{
	my($self) = @_;

	eval { $self->shift while 1 };
	Misc::xdie unless $@ =~ /^(EMPTY)/;
	undef $@;
}

#=# =item peek() : $char
#=# 
#=# Return the oldest character in this MovingLine.
#=# 
#=# Throw an exception starting with "EMPTY" if there are no characters in
#=# this MovingLine.
#=# 
#=# =cut
sub peek
{
	my($self) = @_;

	Misc::xdie "EMPTY peek. " unless @{ $self->{points} } > 0;

	return $self->{points}[0]{item};
}

#=# =item shift() : $char
#=# 
#=# Remove the oldest character from this MovingLine and from the Canvas.
#=# Return the character.
#=# 
#=# Throw an exception starting with "EMPTY" if there are no characters in
#=# this MovingLine.
#=# 
#=# =cut
sub shift
{
	my($self) = @_;

	Misc::xdie "EMPTY shift. " unless @{ $self->{points} } > 0;

	my($point) = CORE::shift @{ $self->{points} };
	$self->{parent}->delete($point->{obj});

	return { -text => $point->{item}, coord => $point->{coord} };
}

#=# =item push($type, %h) : -
#=# 
#=# %h = (
#=# 	coord => $pt,		# starting coordinate
#=# 	rmark => $step,	# retreat mark
#=# 	rcb => $cr_fun,	# retreat callback
#=# 	amark => $step,	# advance mark
#=# 	acb => $cr_fun,	# advance callback
#=# 	-text => $char,	# for 'text'
#=# 	width => $int,		# for 'rectangle'
#=# 	height => $int		# for 'rectangle'
#=# )
#=# 
#=# =cut
{
	my %oktypes = (
		text => 0,
		rectangle => 0
	);
	sub push
	{
		my($self, $type, %rest) = @_;
	
		Misc::xdie "bad type: ``$type''" unless exists $oktypes{$type};
		Misc::xdie "text requires -text"
			if $type eq "text" and ! exists $rest{-text};

		my $curve = $self->{curve};
		my $coord;
		if(exists $rest{coord}) { $coord = delete $rest{coord} }
		elsif($type eq "rectangle")
		{
			$coord = $curve->start;
			$coord->[2] = delete $rest{width} or Misc::xdie "$type needs width";
			$coord->[2] += $coord->[0];
			$coord->[3] = delete $rest{height} or Misc::xdie "$type needs height";
			$coord->[3] += $coord->[1];
		}
		elsif($type eq "text")
		{
			$coord = $curve->start;
			$coord->[1] += 7; # adjust for baseline
		}

		my $point = {
			type => $type,
			item => $rest{-text},
			coord => $coord
		};
		foreach(qw(r a))
		{
			if(exists $rest{"${_}mark"})
			{
				my $c = $_;
				$point->{"${c}mark"} = $curve->getpt(delete $rest{"${c}mark"});
				$point->{"${c}cb"} = delete $rest{"${c}cb"} 
					or Misc::xdie "MARK needs callback. ";
			}
		}

		$point->{obj} = $self->{parent}->create($type, @$coord, %rest);

		CORE::push(@{ $self->{points} }, $point);
	}
}

#=# =item size() : $integer
#=# 
#=# Return the number of characters in this MovingLine.
#=# 
#=# =cut
sub size { scalar @{ $_[0]->{points} } }

#=# =pod
#=# 
#=# =back
#=# 
#=# =cut

1;
###############
package Curve;

use strict;
use Tk;
	
#=# =head1 NAME
#=# 
#=# CurveLine.pm - describes a curve and the points along it
#=# 
#=# =head1 DESCRIPTION
#=# 
#=# This particular class defines a line, but with very little work it
#=# could be changed to a spiral or any other curve.   B<vroom> uses this
#=# class to direct the characters' movement across the screen.
#=# 
#=# =head1 CONSTRUCTOR
#=# 
#=# =over
#=# 
#=# =item new($width, $xpad, $ypad, $percentage) : $CurveLine
#=# 
#=# $width is the length of the line in pixels.   $xpad and $ypad are the
#=# number of pixels to surround the curve with.   $percentage is the
#=# percentage of the line forward() should move a point forward.
#=# 
#=# =back
#=# 
#=# =cut
sub new
{
	my($class, $steps, $xpad, $ypad, $jump) = @_;

	my $self = {
		width => $steps * $jump,
		steps => $steps,
		xpad => $xpad,
		ypad => $ypad,
		jump => $jump
	};

	return bless($self, ref($class) || $class);
}

#=# =head1 METHODS
#=# 
#=# =over
#=# 
#=# =item width() : $number
#=# 
#=# Return the total width this curve wants in pixels including the padding.
#=# 
#=# =cut
sub width { $_[0]->{width} + 2 * $_[0]->{xpad} }

#=# =item cmppt($ar_pt1, $ar_pt2) : $integer
#=# 
#=# Return 1, 0, or -1 if the first point is further on the line, at the
#=# same spot, or behind the second point.
#=# 
#=# =cut
sub cmppt { $_[1]->[0] <=> $_[2]->[0] }

#=# =item getpt($steps) : $ar_pt
#=# 
#=# Return a point that is $steps steps from the starting point of the
#=# curve.   If $steps is negative, then the returned point is abs($steps)
#=# from the start of the curve.
#=# 
#=# =cut
sub getpt
{
	my($self, $steps) = @_;

	Misc::xdie "TOO MANY STEPS. " unless abs($steps) <= $self->{steps};

	my $pt = $self->start;
	$pt->[0] -= ($steps =~ /^-/) # match "-0" too
		? ($self->{steps} + $steps) * $self->{jump}
		: $steps * $self->{jump};

	return $pt;
}

#=# =item height() : $number
#=# 
#=# Return the total height this curve wants in pixels including the padding.
#=# 
#=# =cut
sub height { 2 * $_[0]->{ypad} }

#=# =item start() : $point
#=# 
#=# Return a reference to a two-element array that are the X and Y
#=# coordinates of the starting point of the line.
#=# 
#=# =cut
sub start { [ $_[0]->{width} + $_[0]->{xpad}, $_[0]->{ypad} / 2] }

#=# =item forward($point) : -
#=# 
#=# Modify $point, moves it to the next point on the curve.
#=# 
#=# Throws an exception starting with "END" if the new value of point is
#=# past the end of the curve.
#=# 
#=# =cut
sub forward_point
{
	my($self, $pt) = @_;

	$pt->[0] -= $self->{jump};
	Misc::xdie "END.  " unless $pt->[0] >= $self->{xpad};
}
sub backward_point
{
	my($self, $pt) = @_;

	$pt->[0] += $self->{jump};
	Misc::xdie "START.  " unless $pt->[0] >= $self->{xpad} + $self->{width};
}
sub backward_rectangle
{
	my($self, $pt) = @_;

	$pt->[0] += $self->{jump};
	$pt->[2] += $self->{jump};
	Misc::xdie "START.  " unless $pt->[0] <= $self->{xpad} + $self->{width};
}
sub forward_rectangle
{
	my($self, $pt) = @_;

	$pt->[0] -= $self->{jump};
	$pt->[2] -= $self->{jump};
	Misc::xdie "END.  " unless $pt->[0] >= $self->{xpad};
}

#=# =pod
#=# 
#=# =back
#=# 
#=# =cut

1;
###############
package main;

use Tk;
use Tk::FileDialog ();
use Getopt::Long ();
use IO::File ();

my $End_label = 'end '; # so it doesn't match the index 'end'
my $Start_buf = 10;
my $Characters;
my($Oline, $Iline, $Rline); # Output, Input, Rectangle
my($Speed, $Compete_flag, $Index, $Start_time, $File_dialog);
my($Mbut_help, $Mbut_file, $Input, $Top, $Canvas, $Curve, $Status, $Id);

sub command_start
{
	if(reopen()) { start() } 
	else { butstate('disabled') }
}
sub command_end
{
	$Status->configure(-text => "Stopped sentence, " . forcedone() . ".");
}
sub command_open
{
	butstate('disabled');
	readfile() and butstate('normal');
}
sub command_quit
{
	exit(0);
}
sub command_help
{
	use Pod::Text ();
	my $tl = $Top->Toplevel();
	my $f = $tl->Frame();
	$f->pack(-side => 'top', -fill => 'x');
	$f->Button(-text => 'OK', 
		-command => sub { $tl->destroy }
	)->pack(-side => 'left');
	my $txt = `pod2text $0`;
	$f = $tl->Frame();
	$f->pack(-side => 'top', -fill => 'both');
	$f->Scrolled('Label', 
		-text => $txt, 
		-justify => 'left', 
		-scrollbars => 'se')->pack();

}
sub command_set_speed
{
	my $tl = $Top->Toplevel(-title => 'Set speed');
	my $f1 = $tl->Frame()->pack(-side => 'top');
	$f1->Label(-text => "Characters per minute: ")->pack(-side => 'left');
	my $e = $f1->Entry()->pack(-fill => 'x');

	my $f2 = $tl->Frame()->pack(-side => 'bottom');
	my $b = $f2->Button(-text => 'OK', 
		-command => [ \&_set_speed, $e, $tl])->pack(-side => 'left');
	$f2->Button(-text => 'Cancel', 
		-command => sub { $tl->destroy })->pack(-side => 'left');
}
sub _command_set_speed
{
	my($e, $w) = @_;

	my $n = $e->get;
	if($n =~ /^\s*\d+\s*$/)
	{
		$Speed = 1 / ($n / (60*1000));
		$w->destroy;
	}
	else { mdialog("``$n'' is not a positive integer.") }
}
sub GUIinit()
{
	$Curve = new Curve($Characters, 10, 13, 13);

	$Top = MainWindow->new();
	$File_dialog = $Top->FileDialog();

	my $f1 = $Top->Frame();

	$Canvas = $f1->Canvas(bg => 'white', 
		width => $Curve->width, 
		height => $Curve->height);
	$Canvas->pack(side => 'top', fill => 'x');

	my $mbar = $Top->Frame();
	$mbar->pack(side => 'top', fill => 'x');
	$Mbut_file = $mbar->Menubutton(relief => 'raised', 
		text => "file",
		underline => 0);
	$Mbut_file->pack(side => 'left');
	$Mbut_file->command(-label => 'open', 
		accelerator => 'Alt-o',
		underline => 0,
		command => \&command_open);
	$Top->bind('<Alt-o>', \&command_start);
	$Mbut_file->command(-label => 'start', 
		-accelerator => 'Alt-s',
		-underline => 0,
		-command => \&command_start,
		-state => 'disabled');
	$Top->bind('<Alt-s>', \&command_start);
	$Mbut_file->command(-label => $End_label,
		-command => \&command_end,
		-underline => 0,
		-accelerator => 'Alt-e',
		-state => 'disabled');
	$Top->bind('<Alt-e>', \&command_end);
	$Mbut_file->separator;
	$Mbut_file->checkbutton(-label => "competetive", 
		-variable => \$Compete_flag);
	$Mbut_file->command(-label => 'set speed', -command => \&command_set_speed);
	$Mbut_file->separator;
	$Mbut_file->command(-label => 'quit', 
		-command => \&command_quit,
		-accelerator => 'Alt-q',
		-state => 'normal',
		-underline => 0);
	$Top->bind('<Alt-q>', \&command_quit);

	$f1->pack(side => 'top', fill => 'x');

	my $sframe = $Top->Frame();
	$sframe->pack(-side => 'top', -fill => 'x');
	$sframe->Label(-text => "Status: ")->pack(-side => 'left');
	$Status = $sframe->Label(-text => "");
	$Status->pack(-side => 'left', -fill => 'x');

	$Mbut_help = $mbar->Menubutton(relief => 'raised', 
		text => "help",
		underline => 0);
	$Mbut_help->pack(-side => 'right');
	$Mbut_help->command(-label => 'help', -command => \&command_help);

	$Input = $Top->Entry(width => 25);
	$Input->pack(side => 'top', fill => 'x');

	$Input->bind('<KeyPress>', [\&gotinput, $Input]);
	$Top->bind('all', '<Key-Tab>', '');

	if(@ARGV)
	{
		if(readfile(shift(@ARGV))) { butstate('normal') }
		else { butstate('disabled') }
	}
}

sub butstate
{
	$Mbut_file->entryconfigure('start', -state => $_[0]);
 	$Mbut_file->entryconfigure($End_label, -state => $_[0]);
}

sub mdialog
{
	my($msg) = @_;

	my $tl = $Top->Toplevel();
	$tl->Label(-text=> $msg)->pack(-side => "top");
	$tl->Button(-text => "OK", 
		-command => sub { $tl->destroy })->pack(-side => 'left');
}

sub gotinput()
{
	my($iobj) = @_;

	return unless defined $Iline;
	my(@s) = split //, $iobj->get;
	my($hr);
	while($#s >= $Index && $Iline->size > 0 && $s[$Index] eq $Iline->peek)
	{
		$Index++;
		$hr = $Iline->shift;
		$Oline->push('text', '-text', $hr->{-text}, 'coord', $hr->{coord});
		$Rline->retreat;
	}

	checkdone();
}

sub checkdone
{
	$Status->configure(-text => "Finished sentence, " . forcedone() . ".")
		if($Iline->size == 0 and ! hasmore())
}

sub forcedone
{
	if(defined $Id)
	{
		$Id->cancel;
		undef $Id;
	}
	return wpm($Start_time, 
		time(), 
		substr($Input->get, 0, $Index));
}

sub wpm 
{
	my($start, $end, $line) = @_;

	my $words = 0;
	$words++ while $line =~ /\b\w+\b/g;

	return "WPM:  0" if $words == 0;
	my $time = $end - $start;
	return "WPM:  really fast" if $time == 0;
	my $chars = length $line;
	return sprintf("WPM: %.2f, characters per minute: %.2f, seconds: %i"
		. ", characters: %i",
		$words / ($time / 60),
		$chars / ($time / 60),
		$time,
		$chars);
}

sub start
{
	unless(hasmore())
	{
		$Status->configure(-text => 'empty file.');
		return;
	}

	if(defined $Id)
	{
		$Id->cancel;
		$Id = undef;
	}
	$Input->focus;
 	$Input->delete(0, "end");
	$Status->configure(-text => 'running sentence...');
	$Oline = new MovingLine($Curve, $Canvas);
	$Iline = new MovingLine($Curve, $Canvas);
	$Rline = new MovingLine($Curve, $Canvas);
	$Index = 0;
	my %rargs = (
		'height' => 10,
		'width' => 3,
		'-outline' => 'red'
	);
	if($Compete_flag)
	{
		$Rline->push('rectangle', 
			%rargs,
			rmark => $Start_buf - 1,
			rcb => sub {
				local $@;
				eval { $Oline->advance };
				$Iline->push('text', '-text', getchar()) if hasmore();
				$Iline->advance;
			});
		foreach(1 .. $Start_buf) { feed(0) }
		feed(1);
	}
	else
	{
		$Rline->push('rectangle', 
			%rargs,
			rmark => -2,
			rcb => sub {
				local $@;
				eval { $Oline->advance };
				$Iline->push('text', '-text', getchar()) if hasmore();
				$Iline->advance;
			});
		$Iline->push('text', '-text', ' ');
		$Iline->advance();
		do
		{
			$Iline->push('text', '-text', getchar()) if hasmore();
			eval {
				$Rline->advance;
				$Iline->advance;
			};
		} while(! $@);
		undef $@;
	}
	$Start_time = time();
}

sub feed($)
{
	my($flag) = @_;

	$Iline->push('text', '-text', getchar()) if hasmore();
	eval { 
		local $@;
		$Oline->advance;
	};
	eval {
		$Iline->advance;
		$Rline->advance;
	};
	if($@) 
	{
		$Status->configure(-text => "Sentence was too fast, " 
			. forcedone() . ".");
	}
	elsif($flag) { $Id = $Status->after($Speed, sub { feed(1) }) }
}

{
	my($_Fh, $_Line, $_Next_line, $_File);
	sub reopen { readfile($_File) }
	sub readfile(;$)
	{
		($_File) = @_;

		undef $_Fh;
		$_Line = "";
		$_File = $File_dialog->Show() unless defined $_File;
		return 0 unless $_File;
		$_Fh = new IO::File $_File, "r";
		if(! $_Fh) { mdialog("Can't open ``$_File'', $!") }
		else { $_Next_line = _getline() }

		return defined $_Fh;
	}
	sub getline
	{
		$_Line = $_Next_line;
		return unless defined $_Line;
		$_Next_line = _getline();
		$_Line =~ s/\s$// unless defined $_Next_line;
	}
	sub _getline
	{
		my $line;

		for(;;)
		{
			$line = <$_Fh>;
			last if ! defined $line;
			chomp $line;
			$line =~ s/^\s*//;
			$line =~ s/\s*$/ /;
			$line =~ s/\s+/ /g;
			last unless $line =~ /^\s*$/;
		}

		return $line;
	}
	sub hasmore
	{
		return 0 unless defined $_Line;
		getline() if length $_Line == 0;
		return defined $_Line;
	}
	sub getchar()
	{
		Misc::xdie "EMPTY 1" unless defined $_Line;
		getline() if length $_Line == 0;
		Misc::xdie "EMPTY 2" unless defined $_Line;
		substr($_Line, 0, 1, "");
	}
}

sub getopts
{
	my %opts;
	my $tmp = Getopt::Long::GetOptions(\%opts,
		"characters=i",
		"competitive!",
		"help",
		"speed=i");
	if(exists $opts{help})
	{
		usage();
		exit(0);
	}
	unless($tmp)
	{
		usage();
		exit(1);
	}

	$Compete_flag = exists $opts{competitive} ? $opts{competitive} : 1 ;
	$Characters = exists $opts{characters} ? $opts{characters} : 50 ;
	$Characters = $Start_buf + 1 unless $Characters > $Start_buf;

	# Edit POD if you change this
	if(exists $opts{speed} and $opts{speed} !~ /^\s*\d+\s*$/)
	{
		usage("``$opts{speed}'' is not a non-negative integer.");
		exit(1);
	}
	$Speed = exists $opts{speed} ? ((60*1000) / $opts{speed}) : 1000 ;

	if(@ARGV > 1)
	{
		usage("Can specify at most one input file on the command line.");
		exit(1);
	}
}

sub usage
{
	print STDERR join("\n", @_) . "\n" if @_;
	print STDERR $Usage;
}

#################

getopts();
GUIinit(); 
MainLoop();

