

#! /usr/bin/perl -w

####################################################################
##                             PREFACE
####################################################################

# w3make - a offline web preprocessor for comfortably separating 
# content and layout using XML and XSLT.

# please refer to http://www.skamphausen.de/software/perl/w3make.html
# for more information and newest versions (though I hope this script
# won't need much maintenance ;)

# Copyright (c) 2000 by Stefan Kamphausen
# written by Stefan Kamphausen <mail@skamphausen.de>

# This software was sponsored by
# Novel Science International
# http://www.novelscience.com
# because it was written during my working time there and it's first
# job was to manage YAR (YetAnotherRedesign) of our website.

########
# TODO
########
# * I'd like to have clean-up routines for HXF, (automatically
#   created) DTD (and more?)
#
# * there should be a mode ("devel" or something) in which the COPY
#   keyword in a rule should just symlink the files to reduce stupid
#   copying of large images or the like UPDATE: a time-check is
#   performed everytime which can be overriden by opt_force
#
# * the introduction of dependencies
#
# * every user-assignable name should be tested for shell-special
#   characters if it is used for a filename or dirname or the like.
#   This will need a default replacement for all special chars
#
# * !! the stripping of praeamble from XML-input-files
#
# * !! the link checking after the XSLT process is simply UGLY:
#   - The ROOTLINK keyword should be something more special
#   - relative links should be recognized
#   - the output should be optimized: relative links in cwd
#         others otherwise
#
# * The default in- and outputfilename should be finer grained:
#   For example a PNG should not be named image.html
#
# * The Variable expansion could take place in the outputtree to
#   be able to use variables in the style-sheet. Or maybe I'll pass
#   variables to the style-sheet and make them usable there.
#
# * creation of tarballs

####################################################################
##                             LICENSE
####################################################################
# This program is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
# 02111-1307, USA.

####################################################################
##                              NOTE!
####################################################################
# Please note that this program uses the saxon program by
# Michael H. Kay (see http://users.iclway.co.uk/mhkay) which is a
# java XSLT tool that at the time of this writing implements more
# of the XSLT specification than the Perl Modules XML::Parser,
# XML::DOM and XML::XSLT.
# Unfortunately saxon is written in java and I simply
# hate java. Thus I hope that it will be easily replaced
# when the Perl Modules are ready. 
# My apologies to M.H. Kay, he has done a great job!
# But I don't like java.

####################################################################
##                     PROGRAM IDENTIFICATION
####################################################################
my $program = "w3make";
my $filedate="2001-02-04";
my $fileversion="0.8";
my $copyright = "Copyright 2000 by Stefan Kamphausen";
my $title = "$program $fileversion, $filedate - $copyright\n";

####################################################################
##                             MODULES
####################################################################
use strict;
use vars qw/$opt_xml_dir $root_dir $opt_force $opt_project
  $opt_ask $opt_link $opt_tarball $opt_verbose $opt_help
  $opt_version $opt_locrule $opt_clear/;
use subs qw/msg fatal/;

use Getopt::Long;
use File::Copy;
use File::Path;
use File::Basename;
use FileHandle;
use Cwd;
use POSIX 'strftime';
use HTML::Filter;

####################################################################
##                             GLOBALS
####################################################################
# while debugging...
my $debug=0;
# no outputbuffering
$|++;


my $Usage = <<'END_OF_USAGE';
Usage: w3make [options...] [target]

An offline website preprocessor. It descents a directory looking for
content and layout files and combines them. The processed files are
then copied to a separated place.

  target             If it is given it must be a target in the global
                     rules file. If no target is given the somewhat
                     special target 'all' is used
  -x, --xml_dir=DIR  Specify the input XML directory which will be
                     processed. Your XML tree with the configuration
                     files plus XSL-files and XML-contents lies there.
                     Absolute paths are recognized by the leading
                     slash '/'
                     The default is to use the subdirectory called
                     'xml' in the current working directory
  -p, --project=NAME Create a sample setup for a new project that will
                     be maintained with w3make. Creates some default
                     directories and files.
  -r, --locrule      create a sample local rules file in the current
                     directory
  -c, --clear        Clean up the output directory before recreating it
                     from the scratch. This is useful if you removed a
                     subdirectory from a rules file, which would still
                     be there
  -a, --ask          Ask before creating directories. the default is,
                     not to ask.
  -f, --force        When copying files w3make checks if the source is 
                     newer to avoid useless copying. This option
                     overrides that check.
  -l, --link=CODE    How to handle hyperlinks in your document. w3make
                     will parse the complete outputtree again and replace
                     every 'ROOTLINK' with one of the following according
                     to the code you give:
                     CODE           Replacement
                     web            /
                     abs[olute]     The full path to your document in your
                                    local filesystem
                     rel[ative]     the correct number of '../'s
                     (as a matter of fact, only the first three characters
                      are tested).
  -t, --tarball[=n]  Create a tarball of the created files. The default
                     is to put all the files in the tarball, but you
                     can set this to 'n' if you only want the new files
                     in the archive. NOT_YET_IMPLEMENTED
  -v, --verbose      More output
  --help             Display this help message.
  --version          Display version information.
END_OF_USAGE

my $RootDir = getcwd;
my $runtime = strftime("%Y-%m-%d (%H:%M:%S)",localtime());

## Filenames and Directories
# Name of (local) RulesFile(s)
my $rules_filename = "W3Makefile";
# and the special Global Rules File:
my $rules_global = "W3Makefile.global";
# the default dirs. they are reset optionally in set_global_parms
my $cind = "$RootDir/xml";
my $coutroot = "";
# the current style file (FIXME: need this?)
my $cstyle = "";
# the outputdirectory for the HXF
my $hxf_out = "$RootDir/sites";

## Commands
# FIXME: get command of saxon or error
my $xslparser_name = "saxon";


## Major datastructures
# Hash with all globals except SUBDIRS
my %global_variables = ();
# The Target from commandline or "all"
my $Target = "all";
# Hash for the main_targets (aka LABELS) like HTML or wap or...
# %main_targets{LABEL}{OUTPUTDIR}=subdir_relativ_to_root_or_nothing
# %main_targets{LABEL}{STYLE}=path_to_stylesheet_file MUSTBE!!
# %main_targets{LABEL}{HXF}=Huge_XML_File
# ... and more
# only filled if Target matches LABEL or Target equals "all"
my %main_targets = ();
# pick up all attributes for page and dir elements for
# auto_DTD_creation 
# auto_dtd_atts{dir}{attributname} = 1
my %auto_dtd_atts = ();

## Misc
# current depth in xmltree
# my $depth = 0;
# indentation for each subdir
my $ind = "  ";

####################################################################
##                             OPTIONS
####################################################################
$opt_xml_dir="";
$opt_project="";
$opt_locrule=0;
$opt_clear=0;
$opt_force=0;
$opt_ask=0;
$opt_link="relative";
$opt_tarball="";
$opt_verbose=0;
$opt_help=0;
$opt_version=0;

my $ret = GetOptions( "xml_dir|x=s",
					  "project|p=s",
					  "locrule|r!",
					  "clear|c!",
					  "ask|a!",
					  "force|f!",
					  "tarball|t:s",
					  "link|l=s",
					  "verbose|v!",
					  "help!",
					  "version!"
					);
if (!$ret) {
  die $Usage;
} else {
  # option help
  if ($opt_help) {
	print $Usage;
	exit(0);
  }
  # option version
  if ($opt_version) {
	print $title;
	exit (0);
  }
  if ($opt_project) {
	new_project($opt_project);
	exit (0);
  }
  if ($opt_locrule) {
	new_local_rules();
	exit (0);
  }
  # now the rest of ARGV is the target
  if (@ARGV==1) {
	$Target = $ARGV[0];
  } elsif (@ARGV>1) {
	fatal "sorry, you can't give more than one target\n",
	"try --help for more information\n";
  }
}

####################################################################
##                              MAIN
####################################################################

##
# say a friendly hello
print "This is $title";

##### INIT
# find the XSL parser program
msg "looking for $xslparser_name executable ..";
my $xslparser = qx/which $xslparser_name/;
chomp $xslparser;

unless ($xslparser =~ /$xslparser_name$/) {
  fatal "no XSL parser ($xslparser_name) found\n";
} else {
  msg "OK: $xslparser\n";
}


# parse the main rules_file
# options override rules (this is ensured in that function)
parse_global_rules();

dump_global_targets(%main_targets) if $debug;
dump_global_vars() if $debug;

unless (keys %main_targets) {
  fatal "You need to give a valid target defined in you global rules file\n",
  "see --help for more information";
}
# OK go for it
process_main_targets();

# bye
print "\nReady\n";
exit(0);


####################################################################
##                          SUBROUTINES
####################################################################

########################## PROCESSORS

#### This is kind of main_number_2 since it handles the main targets
sub process_main_targets {

  my ($t,$sty);
  my $underline;
  foreach $t (keys %main_targets) {
	$underline = "=" x (25 + length($t)); 
	print "\nProcessing Main Target \"$t\"\n$underline\n";

	# set output dir
	unless ($main_targets{$t}{OUTPUTDIR}) {
	  fatal "No output directory specified for $t";
	}
	$coutroot = "$RootDir/$main_targets{$t}{OUTPUTDIR}";
	# first clean up (=rm -rf) the dir
	if ($opt_clear) {
	  print "CleanUp: removing previous $coutroot\n";
	  rmtree($coutroot,0,0) or die
		"could not clean up old output directory\n$!";
	}
	# then _all_ resources must come from somewhere else
	# even the pictures should come from inputtree!
	chk_mk_dir($coutroot," ");

	# Set HugeXMLFile
	my $hxf_dir = $main_targets{$t}{HXFDIR} || $hxf_out;
	chk_mk_dir($hxf_dir," ");
	my $chxf = $main_targets{$t}{HXFILE} || "$t.xml";

	# The name of the autogenerated DTD
	my $dtdout = "$hxf_dir/auto_$t.dtd";

	# Create HXF
	# this filehandle can easily be passed to subs
	my $CHXF = new FileHandle ">$hxf_dir/$chxf";
	unless (defined $CHXF) {
	  fatal "couldn't write to $hxf_dir/$chxf\n$!";
	}
	print $CHXF $global_variables{XML},"\n";
	print $CHXF "<!DOCTYPE site SYSTEM \"$dtdout\">\n";	
	# start a site with the outputdir as dir
	print $CHXF "<site name=\"$t\">\n";
	my $dirname = $global_variables{DIR_NAME} || "Home";
	$dirname =~ s/^\s*//;
	print $CHXF " <dir path=\"$coutroot\" name=\"$dirname\" depth=\"0\">\n";
	# Set def!ult sty,e
	$sty = $main_targets{$t}{STYLE};
	if ($sty =~ m(^/)) {
	  $cstyle = $sty;
	} else {
	  $cstyle = "$RootDir/$sty";
	}
	### Process Input Dir (aka main3)
	process_tree($t,$cind,$coutroot,0,$CHXF);

	print $CHXF " </dir>\n</site>\n";
	$CHXF->close;

	print "\nXML tree processing ready.\n\n";
	# Create a DTD from the XML-source-DTD
	# process tree should have been picking up all the
	# attributes for <page> and <dir> on it's way.
	msg "autocreating DTD\n";
	make_dtd($global_variables{DTD},$dtdout);

	# XSLTransformation:
	# process variables in stylefile
	open(STYLE,"<$cstyle") or die
	  "couldn't read current style file $cstyle\n$!";
	my $style_content = "";
	while (<STYLE>) {
	  $style_content .= $_;
	}
	close STYLE;
	# expand global variables in style file
	# FIXME: this is cutnpasted code -> turn to sub
	foreach (keys %global_variables) {
	  $style_content =~ s/\$\($_\)/$global_variables{$_}/g;
	}
	my $c_out_style = "$hxf_dir/" . basename($cstyle);
	open (STYLEOUT,">$c_out_style") or die
	  "couldn't write processed style file $c_out_style\n$!";
	print STYLEOUT $style_content;
	close STYLEOUT;
	# pass HXF and stylefile to saxon...
	# this will probably change in the future
	my $parsecmd = "$xslparser $hxf_dir/$chxf $c_out_style";
	print "Calling Parser \"$xslparser\"\n";
	system $parsecmd and die
	  "Systemcall $parsecmd failed!\n";

	# Linkhandling
	# needs deeper thinking
	if ($t =~ /html/i) { # obviously this is HTML stuff
	                    # this decision should be based on a GlobRule
	  my ($newlink);
	  print "Processing Links. CODE: $opt_link\n";
	  process_links($coutroot,0);
	}
	$underline = "-" x (8 + length($t)); 
	print "\"$t\" done.\n$underline\n";
  }
}

#### Handles the main work for a directory
sub process_tree {
  my ($label,$idir,$odir,$depth,$hxfh) = @_;
  # calc indention
  my $indent = $ind x $depth;
  # FIXME: ist this correct?? One indention for all trees??
  my $xml_indent = $ind;
  # cd w/ check
  unless (-d $idir) {
	fatal "can not read $idir\n";
  }
  chdir($idir) or die "could not change to $idir\n$!";
  print "$indent\\$idir\n";
  # parse local rules
  my @retlist = parse_local_rules($indent);
  # UGLY er what? Can someone please tell me, how to do this better
  my @subdirs = @{$retlist[0]};
  my %loc_targets = %{$retlist[1]};
  my %loc_vars = %{$retlist[2]};
  my %loc_cmds = %{$retlist[3]};
  
  # start a dir in HXF
  my $obasedir = basename($odir);
  my $dirname = $loc_vars{DIR_NAME} || ucfirst($obasedir);
  # sorry, no leading/trailing whitespaces because of filenames
  $dirname =~ s/^\s*//;
  $dirname =~ s/\s*$//;
  unless ($depth==0) {
	print $hxfh $xml_indent, "<dir path=\"$odir\" ",
	"name=\"$dirname\" depth=\"$depth\" ";
	foreach (keys %loc_vars) {
	  if (/^DIR_/) {
		next if /DIR_NAME/; # got this already
		my (undef, $field) = split /_/;
		print $hxfh lc($field),"=\"$loc_vars{$_}\" ";
		# keep the attribute's name for the autoDTD
		unless (defined $auto_dtd_atts{dir}{$field}) {
		  $auto_dtd_atts{dir}{lc($field)} = 1;
		}
	  }
	}
	print $hxfh ">\n";
  }
  
  # Process All Targets (aka main4)
  process_targets($label,\%loc_targets,\%loc_vars,\%loc_cmds,
				  $odir,$idir,$indent,$hxfh);
  
  # process all subdirs
  foreach (@subdirs) {
	# create output mirror
	chk_mk_dir("$odir/$_",$indent);
	# start a dir in HXF
	# and process it
	process_tree($label,$_,"$odir/$_",$depth+1,$hxfh);
  }
  print $hxfh $xml_indent, "</dir>\n" unless ($depth==0);
  # cd ..
  chdir("..") or die "could not change to upper directory\n";
}

#### Takes a list of all targets for a input dir and processes the
# rules 
sub process_targets {
  my ($label,$tl_ref,$vars_ref,$cmds_ref,
	  $odir,$idir,$indent,$hxfh) = @_;
  my %tl = %{$tl_ref};
  my %vars = %{$vars_ref};
  my %cmds = %{$cmds_ref};
  my $ind = "$indent   ";
  my ($l,$t);
  my $infile; # the XML input file to use
  my $outputdir; # hmmm....
  my $outputfile; # now, guess what...
  
  # Main Loop
  foreach $t (keys %tl) { # foreach target
	print "$indent |- $t\n";
	## set INPUT file: 1. rule or 2. all-rule or 3 default
	$infile = $tl{$t}{$label}{INPUT} ||
	  $tl{$t}{all}{INPUT} ||
		"$t.xml";        # default: target_dot_xml
	#unless (-r $infile) {
	#  print "$ind can't read $infile ... skipping\n";
	#  next;
	#}
	msg "$ind INPUT: $infile";
	
	## now handle the ouptutdir (take care for "all" label)
	if (defined $tl{$t}{$label}{OUTPUTDIR}
		|| defined $tl{$t}{all}{OUTPUTDIR}) {
	  
	  my $rulesdir = $tl{$t}{$label}{OUTPUTDIR} ||
		$tl{$t}{all}{OUTPUTDIR};
	  if ($rulesdir eq ".") {
		$outputdir = getcwd;
	  } else { # FIXME: need for absolute/relative path checking?
		$outputdir = $rulesdir;
	  }
	} else {
	  $outputdir = $odir;
	}
	msg "$ind OUTPUTDIR $outputdir";
	
	## now that we got the right dir build the output file
	if (defined($tl{$t}{$label}{OUTPUTFILE}) ||
		defined($tl{$t}{all}{OUTPUTFILE})) {
	  my $rulesfile = $tl{$t}{$label}{OUTPUTFILE} ||
		$tl{$t}{all}{OUTPUTFILE};
	  if ($rulesfile =~ m(^/)) {
		$outputfile = $rulesfile;
	  } else {
		$outputfile = "$outputdir/$rulesfile";
	  }
	} else {
	  # The default rule: take everything of the label until '_'
	  # as the suffix for the target name
	  # this is useful e.g. for html_3.2 ...
	  my ($suffix,undef) = split(/_/,$label);
	  $outputfile = "$outputdir/$t.$suffix";
	}
	msg "$ind OUTPUTFILE $outputfile";

	## do the system calls from the rules
	# FIXME: this is empty for now
	foreach my $cmd (@{$cmds{$t}{$label}}) {
	  msg "$ind EXECUTE: $cmd";
	  system($cmd) and die
		"You gave the following command and it went wrong\n",
		"$cmd\n";
	}
	#print "COMMANDS:\n";
	#print join "\n", @{$cmds{$t}{all}} if defined @{$cmds{$t}{all}};
	#print ">>>\n";
	foreach my $cmd (@{$cmds{$t}{all}}) {
	  msg "$ind EXECUTE: $cmd";
	  system($cmd) and die
		"You gave the following command and it went wrong\n",
		"$cmd\n";
	}
	
	## now either put into HXF or just copy a handmade file
	if (defined $tl{$t}{$label}{COPY} ||
		defined $tl{$t}{all}{COPY}) {
	  my $csrc = $tl{$t}{$label}{COPY} || $tl{$t}{all}{COPY};
	  # FIXME: symlinks if in devel mode (see TODO)
	  if (((stat($csrc))[9] > (stat($outputfile))[9]) |
		 $opt_force > 0) {
		copy($csrc,$outputfile) or die
		  "could not copy $csrc to $outputfile\n$!";
	  } else {
		msg "$ind destination is newer\n";
	  }
	} elsif (defined $tl{$t}{$label}{INPUT} ||
		defined $tl{$t}{all}{INPUT}) {
	  # put it into HXF
	  print $hxfh "$indent<page outputfile=\"$outputfile\">\n";
	  # read the xmlfile
	  open INF, "<$infile" or die
		"could not read $infile\n";
	  my $content = "";
	  while (<INF>) {
		# FIXME !!!!! Hardcoded TopLevelElement in PageDTD
		next if (/<page>/);
		next if (/<\/page>/);
		$content .= $_;
	  }
	  # strip preamble
	  # FIXME: strip XMLDecl and DOCTYPE better!
	  $content =~ s/^\s*<\?xml[^>]*\?>//;
	  $content =~ s/^\s*<!DOCTYPE[^>]*>//;
	  
	  # expand variables in XML file
	  foreach (keys %vars) {
		$content =~ s/\$\($_\)/$vars{$_}/g;
	  }
	  foreach (keys %global_variables) {
		$content =~ s/\$\($_\)/$global_variables{$_}/g;
	  }
	  $content =~ s/\n/\n$indent /g;
	  print $hxfh $content;
	  print $hxfh "$indent</page>\n";
	}
	else {
	  # other code may go here
	  print ".";
	}
  }
}

### Goes thru the output tree and reads all files. They are checked
# for different kind of links (TODO: relative links) which are then
# converted
sub process_links {
  my ($odir,$depth) = @_;
  my ($f,$newlink);
  my $rellink = "./";

  if ($depth > 0) {
	$rellink = "../" x $depth;
  }
  chop($rellink);

  if ($opt_link =~ /^rel/) {
	$newlink = $rellink;
  } elsif ($opt_link =~ /^abs/) {
	$newlink = $coutroot;
  } elsif ($opt_link eq "web") {
	$newlink = "/";
  } else {
	fatal "you gave a wrong option for opt_link: $opt_link\n",
	"choose between 'relative', 'absolute' and 'web'\n";
  }
  print "$odir\n";
  chdir($odir) or die
	"couldn't change to $odir\n$!";
  opendir(DIR,".") or die
	"couldn't open current directory\n$!";
  my @files = readdir(DIR);
  closedir(DIR);
  foreach $f (@files) {
	next if($f eq ".");
	next if($f eq "..");
	if (-d $f) {
	  process_links("$odir/$f",$depth+1);
	} elsif (-f $f && $f =~ /\.html/) {
#	  open F,"<$f" or die
#		"couldn't read $f\n$!";
#	  my $cont = "";
#	  while (<F>) {
#		$cont .= $_;
#	  }
#	  close F or warn "error closing $f\n$!";
#	  my $p = HTML::Filter->new();
#	  $p->parse($cont);
#	  open F,">$f" or die
#		"couldn't write $f\n$!";
#	  print F "HTML"; #filtered_html($p);
#	  close F or warn "error closing $f\n$!";
	  
	  #my $pars = HTML::TokeParser->new($f);
#	  my $tag;
#	  while ($tag = $pars->get_tag) {
#		if ($tag->[0] eq "a") {
#		  print "LINK: $tag->[1]{href}\n";
#		} elsif ($tag->[0] eq "img") {
#		  print "IMG:  $tag->[1]{src}\n";
#		} elsif ($tag->[0] eq "td") {
#		  if (defined $tag->[1]{background}) {
#			print "TDBG: $tag->[1]{background}\n";
#		  }
#		}
#	  }
	  
	  open F,"<$f" or die
		"couldn't read $f\n$!";
	  my @lines = <F>;
	  close F or warn "error closing $f\n$!";
	  foreach (@lines) {
		s/ROOTLINK/$newlink/g;
		s/$coutroot/$newlink/g;
		#s/$rellink/$newlink/g;
	  }
	  open F,">$f" or die
		"could not write $f\n$!";
	  foreach (@lines) {
		print F $_;
	  }
	  close F or warn "error closing $f\n$!";
	} else {
	  #print "don't know how to handle $f\nskipping\n";
	  next;	  
	}	
  }
  chdir "..";
}


########################## PARSER

#### parse the global rules file
# This is special and different from lower rule files since here
# command line options can be set and the directories for the targets
# are set. Furthermore there are special variables...
sub parse_global_rules {
  my ($option);
  
  msg("Reading Global Rule File:\n $RootDir/$rules_global\n");
  my ($target,$val,$key,$dependencies);
  # get the (somwhat cleaned) file
  my @lines = filtered_file("$RootDir/$rules_global");
  # and parse it
  foreach (@lines) {
	if (is_a_variable($_)) {
	  # this is a variable
	  ($key,$val) = split(/=/,$_,2);
	  $key =~ s/\s*//g;
	  # get special OPT_* variables unless the option has been given
	  # on the command line
	  if ($key =~ /^OPT/) {
		s/\s//g;
		$option = lc $key;
		# this line makes that commandline-options override rules file
		unless (eval "\$$option") {
		  msg("\toption set: $option to $val");
		  eval("\$$option=\"$val\"");
		  next;
		}
	  } elsif ($key =~ /_DATE$/) {
		# special Variable: current date
		$global_variables{$key}=strftime("$val",localtime());
	  }
	  # right now: no subdirs, because cind is used
	  else {
		# the usual case is just a normal variable name
		$global_variables{$key}=$val;
	  }
	} else {
	  # no variable
	  unless (/^\s+/) {
		# this is a target-line
		($target,$dependencies) = split(/:/);
		# FIXME: till now we forget dependencies
		$target =~ s/\s//g;
		next;
	  } else {
		next unless want_this_target($target);
		s/^\s+//;
		# Detection Of Keywords
		if (/^style/i) {
		  (undef,$val)=split(/\s+/);
		  $val =~ s/\s//g;
		  $main_targets{$target}{STYLE}=$val;
		} elsif (/^outputdir/i) {
		  (undef,$val)=split(/\s+/);
		  $val =~ s/\s//g;
		  $main_targets{$target}{OUTPUTDIR}=$val;
		} elsif (/^hxfile/i) {
		  $val =~ s/\s//g;
		  $main_targets{$target}{HXFILE}=$val;
		} elsif (/^hxfdir/i) {
		  $val =~ s/\s//g;
		  $main_targets{$target}{HXFDIR}=$val;
		}
	  }
	} # end of no variable
  } # end of @lines

  ## Checks
  # The Vars DTD and XML _must_ be defined
  unless (defined $global_variables{DTD}) {
	fatal "You must give the DTD file of your XML documents",
	"in your global rules";
  }
  unless (defined $global_variables{XML}) {
	fatal "You must give an XML version string in your global rules";
  }
  $global_variables{DTD} =~ s/^\s*//;
  $global_variables{DTD} =~ s/\s*$//;
  $global_variables{XML} =~ s/^\s*//;
  $global_variables{XML} =~ s/\s*$//;
}

sub parse_local_rules {
  my ($indent) = @_;

  # return these three
  my @subs = ();
  my %variables = ();
  my %targets = ();
  my %commands = ();
  
  my ($key,$val,$target,$dependencies,$label,@ruleline);
  # get the rules
  my @lines = filtered_file($rules_filename);
  # and parse
  foreach (@lines) {
	# FIXME (see same place in globals)
	if (is_a_variable($_)) {
	  # this is a variable
	  ($key,$val) = split /\s*=\s*/;
	  if ($key =~ /_DATE$/) {
		# special variable: current date
		$variables{$key}=strftime("$val",localtime());
	  } elsif ($key =~ /SUBDIR/) {
		# special variable: following directories
		foreach (split(/,/,$val)) {
		  s/\s//g;
		  if (/^\//) { # absolut path is a nono
			msg "$_ is an absolute path in the SBUDIR directive\n",
			"this is not allowed. skipping it\n";
		  } else {
			push(@subs,$_);		  
		  }
		}
	  } else {
		# general variables
		$variables{$key} = $val;
	  }
	} # endof Variables
	else {
	  unless (/^\s+/) {
		# this is a target-line
		($target,$dependencies) = split(/:/);
		# FIXME: right now we forget dependencies
		$target =~ s/\s//g;
		next;
	  } else {
		s/^\s+//;
		# Expand Target Variable
		s/\$\(TARGET\)/$target/;
		s/\$\(@\)/$target/;
		# if no label or label eq "all" it should be used for
		# all labels...
		@ruleline = split(/\s+/);
		# FIXME: somewhat ugly, er? Howtodobetter?
		if ($#ruleline == 2) {
		  $label = $ruleline[1];
		  $val = $ruleline[2];
		} elsif ($#ruleline == 1) {
		  $label = "all";
		  $val = $ruleline[1];		  
		} else {
		  # a command
		  $label = "all";
		  $val = $_;
		}
		# Detection Of Keywords
		if (/^outputdir/i) {
		  $val =~ s/\s//g;
		  $targets{$target}{$label}{OUTPUTDIR}=$val;
		} elsif (/^outputfile/i) {
		  $val =~ s/\s//g;
		  $targets{$target}{$label}{OUTPUTFILE}=$val;
		} elsif (/^input/i) {
		  $val =~ s/\s//g;
		  $targets{$target}{$label}{INPUT}=$val;
		} elsif (/^copy/i) {
		  $val =~ s/\s//g;
		  $targets{$target}{$label}{COPY}=$val;
		  $targets{$target}{$label}{OUTPUTFILE}=$val
			unless defined $targets{$target}{$label}{OUTPUTFILE};
		} else {
		  # has no keyword => this is a command
		  # FIXME: commands for each label
		  # for now it's just all-commands
		  # print "COMMAND: $target >>$_<< \n";
		  $targets{$target}{$label}{DEFINED}=1;
		  push(@{$commands{$target}{all}},$_);
		}
	  }
	}
  }
  # Now expand all variables within the rules
  my ($t,$l,$n,$c); # counter
  if (scalar keys %variables > -1) {  
	# all targets
	foreach $t (keys %targets) {
	  # all labels
	  foreach $l (keys %{$targets{$t}}) {
		# all rules
		foreach $n (keys %{$targets{$t}{$l}}) {
		  # all variables
		  foreach (keys %variables) {
			#print "pre : $targets{$t}{$l}{$n}\n";
			$targets{$t}{$l}{$n} =~ s/\$\($_\)/$variables{$_}/g;
			#print "post: $targets{$t}{$l}{$n}\n";
		  }
		}
	  }
	}
	# same expansion in commands
	foreach $t (keys %commands) {
	  foreach $l (keys %{$commands{$t}}) {
		map { foreach $n (keys %variables) {
		  #print "pre : $_\n";
		  $_ =~ s/\$\($n\)/$variables{$n}/g;
		  #$commands{$t}{$l}{$n} =~ s/\$\($_\)/$variables{$_}/g;
		  #print "post: $_\n";
		}
			} @{$commands{$t}{$l}}
	  }
	}
  }
  #dump_local_targets(\%targets);
  
  # Part two of really UGLY returning of array and hashes
  return (\@subs,\%targets,\%variables,\%commands);
}

sub make_dtd {
  my ($dtdin,$dtdout) = @_;
#  open(DTDIN,"<$dtdin") or die
#  	"could not read your DTD ($dtdin)\n$!";
#  my @lines = <DTDIN>;
#  close DTDIN;
  
  open(DTDOUT,">$dtdout") or die
	"could not write generic DTD ($dtdout)\n$!";
  print DTDOUT "<!-- This is a W3Make generic DTD -->\n";
  print DTDOUT "<!ENTITY % xmlsrc SYSTEM \"../$dtdin\">\n";
  print DTDOUT "%xmlsrc;\n\n";

  print DTDOUT "<!ELEMENT site (dir)*>\n";
  print DTDOUT "<!ELEMENT dir (dir | page)*>\n";
  print DTDOUT "<!ATTLIST site name CDATA #REQUIRED>\n";
  print DTDOUT "<!ATTLIST dir \n\tname CDATA #REQUIRED\n",
  "\tpath CDATA #REQUIRED\n",
  "\tdepth CDATA #REQUIRED\n";
  foreach (keys %{$auto_dtd_atts{dir}}) {
	print DTDOUT "\t",$_," CDATA #IMPLIED\n";
  }
  print DTDOUT " >\n";
  close DTDOUT;
}

########################## SAMPLE WRITER

sub new_project {
  my ($name) = @_;

  print "Creating new project \"$name\"...\n";
  print "\tproject directory\n";
  chk_mk_dir($name," ") or die
	"could not create project directory $name\n$!";
  chdir($name) or die
	"could not change to project directory\n$!";
  print "\txml input directory\n";
  chk_mk_dir($cind," ") or die
	"could not create directory xml\n$!";
  print "\tglobal rules file\n";

  open(RULGLO,">$rules_global") or die
	"could not write global rulesfile\n$!";
  print RULGLO "# This is a GlobalRulesFile for $program\n";
  print RULGLO "# created automatically: $runtime by $program\n\n";
  print RULGLO "# MUST BEs:\n";
  print RULGLO "XML = <?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\n";
  print RULGLO "DTD = $name.dtd\n\n";
  print RULGLO "# other variables\n";
  print RULGLO "DIR_NAME = $name\n";
  print RULGLO "BUILD_DATE = %Y-%m-%d (%H.%M.%S)\n\n";
  print RULGLO "# main targets\n";
  print RULGLO "html:\n";
  print RULGLO "\tstyle html.xsl\n";
  print RULGLO "\toutputdir html\n";
  close RULGLO;

  chdir("$cind") or die
	"could not change to directory xml\n$!";
  print "\tlocal rules file in xml-directory\n";
  new_local_rules();
}

sub new_local_rules {
  open(RUL,">$rules_filename") or die
	"could not write new rulesfile";
  print RUL "# This is a RulesFile for $program\n";
  print RUL "# created automatically: $runtime by $program\n\n";
  print RUL "DIR_NAME = Home\n\n";
  print RUL "# Targets:\n";
  print RUL "index:\n";
  print RUL "\tinput html \$(TARGET).xml\n";
  print RUL "\toutputfile html \$(TARGET).html\n";
  close RUL;
}

########################## OTHERS

sub set_global_parms {
  
  # input directory
  if ($opt_xml_dir) {
	if ($opt_xml_dir =~ m(^/)) {
	  $cind = $opt_xml_dir;
	} else {
	  $cind = "$RootDir/$opt_xml_dir";
	}
  }
}


#### Helper Functions

# return true or false according to current target
sub want_this_target {
  my ($t) = @_;
  if ($Target eq "all") {
	return 1;
  } else {
	if ($Target eq $t) {
	  return 1
	}
  }
  return 0;
}
# the detection of variable-setting-lines
sub is_a_variable {
  my ($line) = @_;

  # 1. line optional starts with a letter or _
  # 2. the rest of the variable name is a word-char
  # 3. any amount of whitespaces
  # 4. =
  if ($line =~ /^[a-zA-Z_]?\w+\s*=/) {
	return 1;
  } else {
	return 0;
  }
}
# give out a list of global targets
sub dump_global_targets {
  my (%tl) = @_;
  my $d;
  print "MainTargets:\n";
  foreach $d (keys %tl) {
        print "$d\n";
        foreach (keys %{ $tl{$d} }) {
          print "\t$_: $tl{$d}{$_}\n";
        }
        print "\n";
  }
}
# debug: give out all the non special vars set in global rule file
sub dump_global_vars {
  print "\nThe Global Variables Hash\n",
          "=========================\n";
  foreach (keys %global_variables) {
        print "\t $_ => $global_variables{$_}\n";
  }
  print "=========================\n\n\n";
}
# formatted list of all targets found in a local rulesfile
sub dump_local_targets {
  my ($tref) = @_;
  my %tl = %{ $tref };
  
  my ($t,$l);
  foreach $t (keys %tl) {
	print "TARGET: $t\n";
	foreach $l (keys %{ $tl{$t} }) {
	  print "\tLABEL: $l\n";
	  foreach (keys %{ $tl{$t}{$l}}) {
		print "\t\t$_ $tl{$t}{$l}{$_}\n";
	  }	
	}	
  }
  print "\n";
}
# check for existance of dir or create it
sub chk_mk_dir {
  my ($dir,$indent) = @_;
          
  unless (-d $dir) {
        if ($opt_ask) {
          print "$indent create directory\n$dir ? (y/n) ";
          if (yes_or_no()) {
                fatal "don't create necessary directory $dir";
          }
        }
        msg("$indent creating $dir");
        mkdir($dir,0777) or die
          "couldn't create necessary directory $dir\n$!";
  }
}

# read a file and return a list with all lines that are not
# empty or a comment
sub filtered_file {
  my ($file) = @_;
  open FILE, "<$file" or die
        "could not read $file\n$!";
  my @no_comment_lines = grep(!/^\s*\#/,<FILE>);
  my @no_empty_lines = grep(!/^\s*$/,@no_comment_lines);

  foreach (@no_empty_lines) {
	# strip comments after commands unless # is in quotes
	s/([^\'\"])\#\1.*//;	
	# no EOL, please
	chomp;
  }
  return @no_empty_lines;
}

################## Various output subs
# general output
sub msg {
  if ($opt_verbose) {
	foreach (@_) {
	  print;
	}
	print "\n";
  }
}
# fatal error
sub fatal {
  print "$program fatal: ";
  foreach (@_) {
	print;
  }
  print "\n";
  exit(1);
}
